(* whenjobs
 * Copyright (C) 2012 Red Hat Inc.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License along
 * with this program; if not, write to the Free Software Foundation, Inc.,
 * 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
 *)

open Whenutils

open Unix
open Printf

(* All jobs that are loaded.  Maps name -> [job] structure. *)
let jobs = ref StringMap.empty

(* Map variable names to jobs which depend on that variable.  This
 * gives us a quick way to tell which jobs might need to be reevaluated
 * when a variable is set.
 *)
let dependencies = ref StringMap.empty

(* Current values of variables.  Using the referentially transparent
 * type Map is very useful here because it lets us cheaply keep
 * previous values of variables.
 *)
let variables : variables ref = ref StringMap.empty

(* $HOME/.whenjobs *)
let jobsdir = ref ""

(* Was debugging requested on the command line? *)
let debug = ref false

let esys = Unixqueue.standard_event_system ()

let rec init j d =
  jobsdir := j;
  debug := d;

  Whenlock.create_lock !jobsdir;

  (* Remove old socket if it exists. *)
  let addr = sprintf "%s/socket" !jobsdir in
  (try unlink addr with Unix_error _ -> ());

  ignore (
    Whenproto_srv.When.V1.create_server
      ~proc_reload_file
      ~proc_set_variable
      ~proc_get_variable
      ~proc_get_variable_names
      (Rpc_server.Unix addr)
      Rpc.Tcp (* not TCP, this is the same as SOCK_STREAM *)
      Rpc.Socket
      esys
  )

and proc_reload_file () =
  if !debug then Syslog.notice "remote call: reload_file";

  try reload_file (); `ok
  with Failure err -> `error err

and proc_set_variable (name, value) =
  if !debug then Syslog.notice "remote call: set_variable %s" name;

  let value = variable_of_rpc value in
  variables := StringMap.add name value !variables;

  (* Which jobs need to be re-evaluated? *)
  let jobnames = try StringMap.find name !dependencies with Not_found -> [] in
  let jobs = reevaluate_jobs jobnames in
  run_jobs jobs

and proc_get_variable name =
  if !debug then Syslog.notice "remote call: get_variable %s" name;

  try rpc_of_variable (StringMap.find name !variables)
  with (* all non-existent variables are empty strings *)
    Not_found -> `string_t ""

and proc_get_variable_names () =
  if !debug then Syslog.notice "remote call: get_variable_names";

  (* Only return variables that are non-empty. *)
  let vars = StringMap.fold (
    fun name value xs -> if value <> T_string "" then name :: xs else xs
  ) !variables [] in
  let vars = Array.of_list vars in
  Array.sort compare vars;
  vars

(* Reload the jobs file. *)
and reload_file () =
  let file = sprintf "%s/jobs.cmo" !jobsdir in
  Whenfile.init ();

  let js =
    try
      Dynlink.loadfile file;
      let jobs = Whenfile.get_jobs () in
      Syslog.notice "loaded %d job(s) from %s" (List.length jobs) file;
      jobs
    with
    | Dynlink.Error err ->
      let err = Dynlink.error_message err in
      Syslog.error "error loading jobs: %s" err;
      failwith err
    | exn ->
      failwith (Printexc.to_string exn) in

  (* Set 'jobs' and related global variables. *)
  let () =
    let map = List.fold_left (
      fun map j ->
        let name = j.job_name in
        StringMap.add name j map
    ) StringMap.empty js in
    jobs := map in

  let () =
    let map = List.fold_left (
      fun map j ->
        let deps = dependencies_of_job j in
        let name = j.job_name in
        List.fold_left (
          fun map d ->
            let names = try StringMap.find d map with Not_found -> [] in
            StringMap.add d (name :: names) map
        ) map deps
    ) StringMap.empty js in
    dependencies := map in

  (* Re-evaluate all jobs. *)
  let jobs = reevaluate_jobs (StringMap.keys !jobs) in
  run_jobs jobs

(* Re-evaluate each named job, in a loop until we reach a fixpoint.
 * Return the names of all the jobs that need to be run.
 *)
and reevaluate_jobs jobnames =
  let rec loop set jobnames =
    let set' =
      List.fold_left (
        fun set jobname ->
          let job =
            try StringMap.find jobname !jobs
            with Not_found -> assert false in
          assert (jobname = job.job_name);

          let r, job' = job_evaluate job !variables in
          jobs := StringMap.add jobname job' !jobs;

          if !debug then
            Syslog.notice "evaluate %s -> %b\n" jobname r;

          if r then StringSet.add jobname set else set
      ) set jobnames in
    if StringSet.compare set set' <> 0 then
      loop set' jobnames
    else
      set'
  in
  let set = loop StringSet.empty jobnames in
  StringSet.elements set

and run_jobs jobnames =
  let run_job job =
    Syslog.notice "running %s" job.job_name;
    () (* XXX *)
  in

  List.iter run_job
    (List.map (fun jobname -> StringMap.find jobname !jobs) jobnames)

let main_loop () =
  Unixqueue.run esys
